@@ -40,7 +40,9 @@ import ipywidgets as widgets
 import typing_extensions
 from typing_extensions import Literal, Protocol
 
-from . import _version, patch  # noqa: F401
+import reacton.logging  # noqa: F401  # has sidefx
+
+from . import _version, patch, utils  # noqa: F401
 
 __version__ = _version.__version__
 
@@ -73,7 +75,7 @@ WidgetOrList = Union[widgets.Widget, List[widgets.Widget]]
 EffectCleanupCallable = Callable[[], None]
 EffectCallable = Callable[[], Optional[EffectCleanupCallable]]
 ROOT_KEY = "ROOT::"
-logger = logging.getLogger("react")  # type: ignore
+logger = logging.getLogger("reacton")  # type: ignore
 
 # this will show friendly stack traces
 DEBUG = 0
@@ -149,6 +151,28 @@ def same_component(c1, c2):
     return c1 == c2
 
 
+def _with_tracebacks(e, tracebacks):
+    # copy it, and we need with_traceback for unknown reasons not to cause
+    # an infinite loop
+    e_original = e
+    e = copy.copy(e).with_traceback(e.__traceback__)
+    tb_next = None
+
+    # last item is the top of the stack
+    for tb in tracebacks:
+        # make a copy, so we do not mutate the original traceback
+        tb = TracebackType(tb_next=tb_next, tb_frame=tb.tb_frame, tb_lasti=tb.tb_lasti, tb_lineno=tb.tb_lineno)
+        tb_next = tb
+
+    if TRACEBACK_ORIGINAL:
+        e = e.with_traceback(tb_next)
+        e_original.__cause__ = e
+        return e_original
+    else:
+        e = e.with_traceback(tb_next)
+        return e
+
+
 class ComponentCreateError(RuntimeError):
     def __init__(self, rich_traceback):
         super().__init__(rich_traceback)
@@ -283,7 +307,8 @@ class Element(Generic[W]):
         assert rc.context is self._current_context, f"Context change from {self._current_context} -> {rc.context}"
         assert rc.context is not None
         ca = rc.container_adders.pop()
-        self.add_children(ca.collect())
+        collected = ca.collect()
+        self.add_children(collected)
 
     def add_children(self, children):
         if self.child_prop_name not in self.kwargs:
@@ -381,7 +406,10 @@ class Element(Generic[W]):
     def _remove_widget_event_listener(self, widget: widgets.Widget, name: str, callback: Callable):
         target_name = name[3:]
         on_change = self._callback_wrappers[callback]
-        widget.unobserve(on_change, target_name)
+        try:
+            widget.unobserve(on_change, target_name)
+        except ValueError:
+            logger.error("Could not remove event listener %r from %r", name, widget)
 
 
 class Value(Generic[V], Protocol):
@@ -429,24 +457,27 @@ class ValueElement(Generic[W, V], Element[W]):
 FuncT = TypeVar("FuncT", bound=Callable[..., Element])
 
 
-def find_children(el):
-    children = set()
-    if not isinstance(el.kwargs, dict):
-        raise RuntimeError("keyword arguments for {el} should be a dict, not {el.kwargs}")
-    for arg in list(el.kwargs.values()) + list(el.args):
-        if isinstance(arg, Element):
-            children.add(arg)
-        elif isinstance(arg, (tuple, list)):
-            for child in arg:
-                if isinstance(child, Element):
-                    children.add(child)
-                    children |= find_children(child)
-        elif isinstance(arg, dict):
-            for child in arg.values():
-                if isinstance(child, Element):
-                    children.add(child)
-                    children |= find_children(child)
-    return children
+def find_elements(value: Union[Element, List, Tuple, Dict]) -> Set[Element]:
+    if isinstance(value, Element):
+        el = value
+        elements = {el}
+        if not isinstance(el.kwargs, dict):
+            raise RuntimeError("keyword arguments for {el} should be a dict, not {el.kwargs}")
+        elements |= find_elements(el.args)
+        elements |= find_elements(el.kwargs)
+        return elements
+    elif isinstance(value, (tuple, list)):
+        elements = set()
+        for child in value:
+            if isinstance(child, (Element, tuple, list, dict)):
+                elements |= find_elements(child)
+        return elements
+    elif isinstance(value, dict):
+        elements = set()
+        for child in value.values():
+            if isinstance(child, (Element, tuple, list, dict)):
+                elements |= find_elements(child)
+        return elements
 
 
 class ContainerAdder(Generic[W]):
@@ -461,7 +492,7 @@ class ContainerAdder(Generic[W]):
     def collect(self):
         children = set()
         for el in self.created:
-            children |= find_children(el)
+            children |= find_elements(el) - {el}
         top_level = [k for k in self.created if k not in children]
         return top_level
 
@@ -527,17 +558,22 @@ def component(obj: FuncT, mime_bundle=...) -> FuncT:
     ...
 
 
+@overload
+def component(obj: Callable[P, None], mime_bundle=...) -> Callable[P, Element]:
+    ...
+
+
 # it is actually this...
 # def component(obj: Union[Type[widgets.Widget], FuncT]) -> Union[ComponentWidget, ComponentFunction[FuncT]]:
 # but casting to FuncT gives much better type hints (e.g. argument types checks etc)
 
 
-def component(obj: FuncT = None, mime_bundle: Dict[str, Any] = mime_bundle_default):
-    def wrapper(obj: FuncT) -> FuncT:
+def component(obj: Union[Callable[P, None], FuncT] = None, mime_bundle: Dict[str, Any] = mime_bundle_default):
+    def wrapper(obj: Union[Callable[P, None], FuncT]) -> FuncT:
         if isclass(obj) and issubclass(obj, widgets.Widget):
             return cast(FuncT, ComponentWidget(widget=obj, mime_bundle=mime_bundle))
         else:
-            return cast(FuncT, ComponentFunction(f=obj, mime_bundle=mime_bundle))
+            return cast(FuncT, ComponentFunction(f=cast(FuncT, obj), mime_bundle=mime_bundle))
 
     if obj is not None:
         return wrapper(obj)
@@ -611,7 +647,7 @@ def use_state(initial: T, key: str = None, eq: Callable[[Any, Any], bool] = None
 
     This function can only be called from a component function.
 
-    The value rturns the current state (which equals initial at the first render call).
+    The value returns the current state (which equals initial at the first render call).
     Or the value that was last
 
     Subsequent
@@ -969,7 +1005,7 @@ class _RenderContext:
         else:
             memo = self.context.memo[self.context.memo_index]
             value, dependencies_previous = memo
-            if dependencies_previous == dependencies:
+            if utils.equals(dependencies_previous, dependencies):
                 logger.info("Got memo hit = %r for index %r (debug-name: %r)", memo, self.context.memo_index, name)
             else:
                 logger.info("Replace memo with = %r for index %r (debug-name: %r)", memo, self.context.memo_index, name)
@@ -988,6 +1024,8 @@ class _RenderContext:
             self.context.state[key] = initial
             if isinstance(initial, (list, dict, set)):
                 self.context.state_metadata[key] = len(initial)
+            elif utils.isinstance_lazy(initial, "pandas.DataFrame"):
+                self.context.state_metadata[key] = utils.dataframe_fingerprint(initial)
             logger.info("Initial state = %r for key %r (%r)", initial, key, id(self.context))
             return initial, self.make_setter(key, self.context, eq)
         else:
@@ -1001,28 +1039,46 @@ class _RenderContext:
                 value = value(context.state[key])
             logger.info("Set state = %r for key %r (previous value was %r) (%r)", value, key, context.state[key], id(self.context))
 
-            if context.state[key] is value and isinstance(value, (list, dict, set)):
-                if context.state_metadata[key] != len(value):
-                    warn(
-                        "You are setting the state with the same object, this will not trigger a rerender. "
-                        f"We noticed the length of {value} changed compared to the previous time it was set. Are you mutating an existing state object? "
-                        "A common mistake is appending to a list, mutating a dict or set, etc.",
-                        UserWarning,
-                        stacklevel=2,
-                    )
-
-            should_update = not eq(context.state[key], value) if eq is not None else context.state[key] != value
+            should_update = False
+            new_metadata = None
+            if eq is None:
+                if context.state[key] is value and isinstance(value, (list, dict, set)):
+                    new_metadata = len(value)
+                    if context.state_metadata[key] != new_metadata:
+                        warn(
+                            "You are setting the state with the same object, this will usually not trigger a rerender. "
+                            f"The length of {value} changed compared to the previous time it was set. Are you mutating an existing state object? "
+                            "A common mistake is appending to a list, mutating a dict or set, etc.",
+                            UserWarning,
+                            stacklevel=2,
+                        )
+                        should_update = True
+                if context.state[key] is value and utils.isinstance_lazy(value, "pandas.DataFrame"):
+                    new_metadata = utils.dataframe_fingerprint(value)
+                    if context.state_metadata[key] != new_metadata:
+                        warn(
+                            "You are setting the state with the dataframe, this will usually not trigger a rerender. "
+                            "We noticed the ids of the dataframe series are changed. Are you mutating an dataframe? "
+                            "Consider making a copy of the dataframe.",
+                            UserWarning,
+                            stacklevel=2,
+                        )
+                        should_update = True
+            equals = eq or utils.equals
+            should_update = not equals(context.state[key], value) or should_update
 
             # if a cleanup during close trigger a state update in a different component, we want to ignore that
             # otherwise we get a deadlock
             if self._closing:
-
                 should_update = False
 
             if should_update:
                 context.state[key] = value
-                if isinstance(value, (list, dict, set)):
-                    context.state_metadata[key] = len(value)
+                if context.state[key] is value and isinstance(value, (list, dict, set)) and new_metadata is None:
+                    new_metadata = len(value)
+                if context.state[key] is value and utils.isinstance_lazy(value, "pandas.DataFrame") and new_metadata is None:
+                    new_metadata = utils.dataframe_fingerprint(value)
+                context.state_metadata[key] = new_metadata
                 # TODO: enable
                 # context.needs_render = True
                 if self._rerender_needed is False:
@@ -1111,6 +1167,8 @@ class _RenderContext:
                         # we started the rendering loop (main_render_phase is True), so we keep going
                         # but if an exception bubbled up, we should stop
                         while self._rerender_needed and not self.context_root.exceptions_children:
+                            if render_counts > 50:
+                                raise RuntimeError(f"Too many renders triggered, your render loop does not stop, last reason: {self._rerender_needed_reason}")
                             logger.info("Entering nested render phase: %r", self._rerender_needed_reason)
                             self._rerender_needed = False
                             self._rerender_needed_reason = None
@@ -1123,8 +1181,6 @@ class _RenderContext:
                             logger.info("Render done: %r %r", self._rerender_needed, self._rerender_needed_reason)
                             assert self.context is self.context_root
                             render_counts += 1
-                            if render_counts > 50:
-                                raise RuntimeError("Too many renders triggered, your render loop does not stop")
                         logger.debug("Render phase resulted in (next) elements:")
                         for el in self._shared_elements_next:
                             logger.debug("\t%r %x", el, id(el))
@@ -1183,21 +1239,8 @@ class _RenderContext:
                     # construct a fake traceback (showing how the elements were constructed)
                     if not self.tracebacks:
                         raise
-                    # copy it, and we need with_traceback for unknown reasons not to cause
-                    # an infinite loop
-                    e_original = copy.copy(e).with_traceback(e.__traceback__)
-                    tb_next = None
-
-                    # last item is the top of the stack
-                    for tb in self.tracebacks:
-                        # make a copy, so we do not mutate the original traceback
-                        tb = TracebackType(tb_next=tb_next, tb_frame=tb.tb_frame, tb_lasti=tb.tb_lasti, tb_lineno=tb.tb_lineno)
-                        tb_next = tb
-
-                    if TRACEBACK_ORIGINAL:
-                        raise e.with_traceback(tb_next) from e_original
-                    else:
-                        raise e.with_traceback(tb_next)
+                    e = _with_tracebacks(e, self.tracebacks)
+                    raise e
                 else:
                     raise
 
@@ -1208,9 +1251,12 @@ class _RenderContext:
 
         exceptions = [*self.context.exceptions_children, *self.context_root.exceptions_self]
         if exceptions:
+            exc = exceptions[0]
+            if DEBUG:
+                exc = _with_tracebacks(exc, self.tracebacks)
+
             if self.handle_error:
                 logger.info("Exception occurred, rendering error message")
-                exc = exceptions[0]
                 if exc.__traceback__ is None:
                     value = "Exception occurred, but no traceback available"
                 else:
@@ -1222,7 +1268,7 @@ class _RenderContext:
 
                 return self.render(w.HTML(value="<pre>" + value + "</pre>"), self.container)
             else:
-                raise exceptions[0]
+                raise exc
         return widget
 
     def _render(self, element: Element, default_key: str, parent_key: str):
@@ -1331,7 +1377,16 @@ class _RenderContext:
                 # back the root element
                 root_element: Optional[Element] = None
                 try:
-                    root_element = el.component.f(*el.args, **el.kwargs)
+                    if _default_container is not None:
+                        with _default_container() as container:
+                            root_element = el.component.f(*el.args, **el.kwargs)
+                        if root_element is None:
+                            if len(container.kwargs["children"]) == 1:
+                                root_element = container.kwargs["children"][0]
+                            else:
+                                root_element = container
+                    else:
+                        root_element = el.component.f(*el.args, **el.kwargs)
                 except BaseException as e:
                     if DEBUG:
                         # we might be interested in the traceback inside the call...
@@ -1390,6 +1445,13 @@ class _RenderContext:
                     # child does not handle exceptions, so bubble up
                     self.context.exceptions_children.extend(context.exceptions_self)
                     self.context.exceptions_children.extend(context.exceptions_children)
+                if self.context.exceptions_self or self.context.exceptions_children:
+                    if not self._rerender_needed:
+                        # this happens when an exception was added from an event handler
+                        # this means no exception was raised during the render phase
+                        # but we still need to rerender, until someone catches the exception
+                        self._rerender_needed_reason = "Exception ocurred during during event handler"
+                        self._rerender_needed = True
 
             assert context is not None
 
@@ -1405,8 +1467,6 @@ class _RenderContext:
         logger.debug("Reconsolidate: (%s,%s) %r", parent_key, key, el)
         context = self.context
         assert context is not None
-        assert not context.exceptions_children, f"Exceptions found in reconciliation phase: {context.exceptions_children} for context: {context!r}"
-        assert not context.exceptions_self, f"Exceptions found in reconciliation phase: {context.exceptions_self} for context: {context!r}"
 
         el_prev = context.elements.get(key)
 
@@ -1480,7 +1540,7 @@ class _RenderContext:
                         if effect.next:
                             # if we have a next, it means that effect itself is executed
                             # TODO: custom equals
-                            if effect.next.dependencies is not None and effect.dependencies == effect.next.dependencies:
+                            if effect.next.dependencies is not None and utils.equals(effect.dependencies, effect.next.dependencies):
                                 logger.info("No need to add effect, dependencies are the same (%r)", effect.dependencies)
                                 # not needed, just remove the reference
                                 effect.next = None
@@ -1693,12 +1753,16 @@ class _RenderContext:
                 self._visit_children(el, key, parent_key, self._remove_element)
             try:
                 self.context = child_context = context.children[key]
+                # we are removing the element, so we want to handle the
+                # exceptions any more, only if the cleanup fails
+                child_context.exceptions_self = []
+                child_context.exceptions_children = []
 
                 for effect_index, effect in enumerate(self.context.effects):
                     try:
                         effect.cleanup()
                     except BaseException as e:
-                        context.exceptions_self.append(e)
+                        child_context.exceptions_self.append(e)
                         self._rerender_needed_reason = "Exception ocurred during effect"
                         self._rerender_needed = True
                 assert self.context.root_element is not None
@@ -1707,6 +1771,10 @@ class _RenderContext:
             finally:
                 # restore context
                 self.context = context
+            if child_context.exceptions_self or child_context.exceptions_children and not child_context.exception_handler:
+                # child does not handle exceptions, so bubble up
+                context.exceptions_children.extend(child_context.exceptions_self)
+                context.exceptions_children.extend(child_context.exceptions_children)
             del context.children[key]
         else:
             self._visit_children(el, key, parent_key, self._remove_element)
@@ -1799,6 +1867,7 @@ def display(el: Element, mime_bundle: Dict[str, Any] = mime_bundle_default):
     import IPython.display  # type: ignore
 
     box = widgets.VBox(_view_count=0)
+    el = _wrap(el, jupyter_decorator_components)
     widget, rc = render(el, container=box)
     displayed = False
 
@@ -1823,6 +1892,17 @@ def display(el: Element, mime_bundle: Dict[str, Any] = mime_bundle_default):
         widget._handle_displayed()
 
 
+# list of decorators that wrap an element when displayed
+# in a jupyter environment, like notebook, or lab via _ipython_display_
+jupyter_decorator_components: List[Callable[..., Element]] = []
+
+
+def _wrap(el: Element, decorators: List[FuncT]) -> Element:
+    for c in jupyter_decorator_components:
+        el = c(children=[el])
+    return el
+
+
 def make(el: Element, handle_error: bool = True):
     hbox = widgets.VBox(_view_count=0)
     _, rc = render(el, hbox, "children", handle_error=handle_error)
@@ -1830,6 +1910,7 @@ def make(el: Element, handle_error: bool = True):
 
 
 _last_interactive_vbox = None
+_default_container: Optional[Callable[..., Element]] = None
 
 
 def component_interactive(static=None, **kwargs):
