@@ -8,10 +8,12 @@ import ipyvuetify
 import ipywidgets
 import ipywidgets as widgets
 import numpy as np
+import pandas as pd
 import pytest
 import traitlets
 
 import reacton as react
+import reacton.core
 from reacton.core import component, use_effect
 
 from . import bqplot  # noqa: F401
@@ -89,6 +91,10 @@ def Container(request):
     return dict(ContainerWidget=w.HBox, ContainerFunction=ContainerFunction)[request.param]
 
 
+Container1 = Container
+Container2 = Container
+
+
 def test_internals():
     @react.component
     def Child():
@@ -554,6 +560,24 @@ def test_state_outside():
     checkbox.close()
 
 
+def test_children_collect():
+    class SomeContainer(widgets.Widget):
+        v_slots = traitlets.Any(widgets.Widget)
+
+    @react.component
+    def Test():
+        with w.HBox() as main:
+            button = w.Button(description="Button")
+            SomeContainer.element(v_slots=[{"name": "extension", "children": button}])
+        return main
+
+    vbox, rc = react.render(Test(), handle_error=False)
+    # should only have SomeContainer as child
+    assert len(rc.find(widgets.HBox).widget.children) == 1
+
+    rc.close()
+
+
 def test_children():
     clear()
     # hbox = widgets.HBox()
@@ -1682,10 +1706,7 @@ def test_exception_handler_in_event_widget():
     clear = None
 
     @react.component
-    def Container():
-        nonlocal clear, exception
-        exception, clear = react.use_exception()
-
+    def Button():
         def on_click():
             raise Exception("in event handler: on_click")
 
@@ -1694,11 +1715,23 @@ def test_exception_handler_in_event_widget():
 
         return w.Button(on_click=on_click, on_description=on_description)
 
+    @react.component
+    def Container():
+        nonlocal clear, exception
+        exception, clear = react.use_exception()
+        # breakpoint()
+        if exception:
+            return w.IntSlider()
+        else:
+            return Button()
+
     box, rc = react.render(Container(), handle_error=False)
     assert clear is not None
     rc.find(ipywidgets.Button).widget.click()
     assert exception is not None
     assert str(exception) == "in event handler: on_click"
+    rc.find(ipywidgets.IntSlider).assert_not_empty()
+    clear()
     rc.find(ipywidgets.Button).widget.description = "new description"
     assert str(exception) == "in event handler: on_description"
     # because the Handler failed to catch the error, the core library does it
@@ -1788,6 +1821,32 @@ def test_recover_exception_in_reconcilliate():
     rc.close()
 
 
+def test_recover_exception_in_cleanup_child(Container1, Container2):
+    set_fail = None
+
+    @react.component
+    def Test():
+        nonlocal set_fail
+        fail, set_fail = react.use_state(False)
+
+        def some_effect():
+            def cleanup():
+                # breakpoint()
+                raise Exception("fail")
+
+            return cleanup
+
+        react.use_effect(some_effect, [fail])
+        return w.IntSlider()
+
+    box, rc = react.render(Container1(children=[Test()]), handle_error=False)
+    assert set_fail is not None
+    assert rc.find(ipywidgets.IntSlider)
+    with pytest.raises(Exception):
+        rc.render(Container1(children=[Container2()]))
+    rc.close()
+
+
 def test_recover_exception_in_cleanup():
     set_fail = None
 
@@ -2362,14 +2421,123 @@ def test_mutate_warning():
         items, set_items = react.use_state([1, 2, 3])
 
         def add_item():
-            items.append(len(items))
+            items.append(len(items) + 1)
             set_items(items)
 
-        return w.Button(on_click=add_item)
+        return w.Button(description=str(items), on_click=add_item)
 
     box, rc = react.render(Test())
     with pytest.warns(UserWarning, match="mutating"):
         rc.find(widgets.Button).widget.click()
+    # even if we get a warning, it should rerender
+    assert rc.find(widgets.Button).widget.description == "[1, 2, 3, 4]"
+    rc.close()
+
+
+def test_set_state_with_dataframe():
+    df1 = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
+    df2 = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 7]})
+    set_df = None
+
+    @react.component
+    def Test():
+        nonlocal set_df
+        df, set_df = react.use_state(cast(pd.DataFrame, None))
+        # set_df(df))
+        return w.Button(description=str(df))
+
+    box, rc = react.render(Test(), handle_error=False)
+    assert set_df is not None
+    assert rc.find(widgets.Button).widget.description == "None"
+    assert rc.render_count == 1
+    set_df(df1)
+    assert rc.render_count == 2
+    # should not rerender
+    set_df(df1)
+    assert rc.render_count == 2
+    assert rc.find(widgets.Button).widget.description == str(df1)
+    set_df(df2)
+    assert rc.render_count == 3
+    assert rc.find(widgets.Button).widget.description == str(df2)
+
+    with pytest.warns(UserWarning, match="mutating"):
+        # mutate the dataframe
+        df2["b"] = df2["b"] + 1
+        # should trigger the warning
+        set_df(df2)
+
+    rc.close()
+
+
+def test_use_memo_dataframe():
+    df1 = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
+    df2 = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 7]})
+    set_df = None
+    use_memos = 0
+
+    @react.component
+    def Test():
+        nonlocal set_df
+        nonlocal use_memos
+        df, set_df = react.use_state(cast(pd.DataFrame, None))
+
+        def increase_use_memos():
+            nonlocal use_memos
+            use_memos += 1
+
+        react.use_memo(increase_use_memos, [df])
+        return w.Button(description=str(df))
+
+    box, rc = react.render(Test(), handle_error=False)
+    assert set_df is not None
+    assert rc.find(widgets.Button).widget.description == "None"
+    assert rc.render_count == 1
+    assert use_memos == 1
+    set_df(df1)
+    assert use_memos == 2
+    # should not trigger use_memo
+    set_df(df1)
+    assert use_memos == 2
+    set_df(df2)
+    assert use_memos == 3
+
+    # TODO: mutation detection in use_memo
+    rc.close()
+
+
+def test_use_effect_dataframe():
+    df1 = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
+    df2 = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 7]})
+    set_df = None
+    use_effects = 0
+
+    @react.component
+    def Test():
+        nonlocal set_df
+        nonlocal use_effects
+        df, set_df = react.use_state(cast(pd.DataFrame, None))
+
+        def increase_use_effects():
+            nonlocal use_effects
+            use_effects += 1
+
+        react.use_effect(increase_use_effects, [df])
+        return w.Button(description=str(df))
+
+    box, rc = react.render(Test(), handle_error=False)
+    assert set_df is not None
+    assert rc.find(widgets.Button).widget.description == "None"
+    assert rc.render_count == 1
+    assert use_effects == 1
+    set_df(df1)
+    assert use_effects == 2
+    # should not trigger use_effect
+    set_df(df1)
+    assert use_effects == 2
+    set_df(df2)
+    assert use_effects == 3
+
+    # TODO: mutation detection in use_effect
     rc.close()
 
 
@@ -2505,3 +2673,49 @@ def test_value_component():
     rc.find(widgets.IntSlider).widget.value = 43
     assert value.value == 43
     rc.close()
+
+
+def test_jupyter_decorator():
+    context = react.create_context(1)
+
+    @react.component
+    def Decorator(children=[]):
+        context.provide(42)
+        return children[0]
+
+    reacton.core.jupyter_decorator_components.append(Decorator)
+    try:
+
+        @react.component
+        def Test():
+            value = react.use_context(context)
+            return w.Button(description=str(value))
+
+        react.display(Test())
+        assert react.core.local.last_rc is not None
+        react.core.local.last_rc.find(widgets.Button).assert_matches(description="42")
+        react.core.local.last_rc.close()
+    finally:
+        reacton.core.jupyter_decorator_components.pop()
+
+
+def test_implicit_component():
+    reacton.core._default_container = w.VBox
+
+    @react.component
+    def Test(a: int):
+        w.Button(description="hoeba: " + str(a))
+
+    box, rc = react.render(Test(1), handle_error=False)
+    assert rc.find(widgets.Button).widget.description == "hoeba: 1"
+    rc.close()
+
+    @react.component
+    def Test2(a: int):
+        w.Button(description="hoeba: " + str(a))
+        w.Button(description="hoeba: " + str(a + 1))
+
+    box, rc = react.render(Test2(1), handle_error=False)
+    assert rc.find(widgets.Button)[0].widget.description == "hoeba: 1"
+    assert rc.find(widgets.Button)[1].widget.description == "hoeba: 2"
+    rc.close()
